/* Copyright (C) 2020  C. McEnroe <june@causal.agency>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <err.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>

#include "server.h"

bool contextsPublic;
int contextsRecent = 500;

const char *ContextsMOTDQuery = SQL(
	SELECT motd
	FROM motds
	WHERE network = :network
	ORDER BY time DESC
	LIMIT 1;
);

const char *ContextsQuery = SQL(
	WITH recentEvents AS (
		SELECT time, context
		FROM events
		ORDER BY event DESC
		LIMIT :recent
	), activeContexts AS (
		SELECT name, query
		FROM contexts
		JOIN recentEvents USING (context)
		WHERE network = :network AND query <= NOT :public
		GROUP BY context
		ORDER BY max(time) DESC
	), allContexts AS (
		SELECT name, query
		FROM contexts
		WHERE network = :network AND query <= NOT :public
		ORDER BY query, name COLLATE NOCASE
	)
	SELECT name, query, 1 FROM activeContexts
	UNION ALL
	SELECT name, query, 0 FROM allContexts;
);

enum kcgi_err contextsPage(struct kreq *req) {
	if (!req->fieldmap[Network]) return httpFail(req, KHTTP_400);
	const char *network = req->fieldmap[Network]->parsed.s;

	enum kcgi_err error = 0
		|| httpHead(req, KHTTP_200, KMIME_TEXT_HTML)
		|| khttp_body(req);
	if (req->method == KMETHOD_HEAD) return error;

	struct khtmlreq html;
	error = error
		|| khtml_open(&html, req, 0)
		|| htmlHead(&html, req, network)
		|| htmlNav(&html, req);
	if (error) return error;

	sqlite3_reset(stmt.contexts);
	dbBindText(stmt.contexts, ":network", network);
	dbBindInt(stmt.contexts, ":recent", contextsRecent);
	dbBindInt(stmt.contexts, ":public", contextsPublic);

	enum State {
		None,
		Active,
		Channels,
		Queries,
	} state = None;
	const char *Headings[] = { NULL, "Active", "Channels", "Queries" };

	int result;
	while (SQLITE_ROW == (result = sqlite3_step(stmt.contexts))) {
		int i = 0;
		const char *context = sqlite3_column_text(stmt.contexts, i++);
		bool query = sqlite3_column_int(stmt.contexts, i++);
		bool active = sqlite3_column_int(stmt.contexts, i++);

		enum State prev = state;
		state = (active ? Active : query ? Queries : Channels);
		if (state != prev) {
			error = 0
				|| khtml_closeelem(&html, 1)
				|| khtml_elem(&html, KELEM_H2)
				|| khtml_puts(&html, Headings[state])
				|| khtml_closeelem(&html, 1)
				|| khtml_elem(&html, KELEM_UL);
			if (error) return error;
		}

		char *href = khttp_urlpart(
			NULL, NULL, Pages[Events],
			Keys[Network].name, network,
			Keys[Context].name, context,
			NULL
		);
		if (!href) err(EX_OSERR, "khttp_urlpart");
		error = 0
			|| khtml_elem(&html, KELEM_LI)
			|| khtml_attr(&html, KELEM_A, KATTR_HREF, href, KATTR__MAX)
			|| khtml_puts(&html, context)
			|| khtml_closeelem(&html, 2);
		free(href);
		if (error) return error;
	}
	if (result != SQLITE_DONE) errx(EX_SOFTWARE, "%s", sqlite3_errmsg(db));

	error = khtml_closeelem(&html, 1);
	if (error) return error;

	sqlite3_reset(stmt.motd);
	dbBindText(stmt.motd, ":network", network);

	result = sqlite3_step(stmt.motd);
	if (result == SQLITE_ROW) {
		error = 0
			|| khtml_elem(&html, KELEM_H2)
			|| khtml_puts(&html, "MOTD")
			|| khtml_closeelem(&html, 1)
			|| khtml_elem(&html, KELEM_PRE)
			|| htmlIRC(&html, sqlite3_column_text(stmt.motd, 0))
			|| khtml_closeelem(&html, 1);
		if (error) return error;
		result = sqlite3_step(stmt.motd);
	}
	if (result != SQLITE_DONE) errx(EX_SOFTWARE, "%s", sqlite3_errmsg(db));

	return htmlFooter(&html, req) || khtml_close(&html);
}
