/* Copyright (C) 2020  C. McEnroe <june@causal.agency>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <err.h>
#include <inttypes.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <time.h>

#include "server.h"

int eventsGap = 3600;
int eventsOverlap = 30;
int eventsLimit = 50;

static const char *timestamp(time_t time) {
	static char stamp[sizeof("0000-00-00T00:00:00")];
	strftime(stamp, sizeof(stamp), "%FT%T", gmtime(&time));
	return stamp;
}

static enum kcgi_err
dateForm(struct khtmlreq *html, struct kreq *req, const char *current) {
	if (req->fieldmap[Export]) return KCGI_OK;
	struct tm tm = {0};
	if (!strptime(current, "%F", &tm)) {
		tm = *gmtime(&(time_t) { time(NULL) });
	}
	char date[sizeof("0000-00-00")];
	strftime(date, sizeof(date), "%F", &tm);
	return 0
		|| khtml_attr(
			html, KELEM_FORM,
			KATTR_METHOD, "get",
			KATTR_ACTION, Pages[Events],
			KATTR__MAX
		)
		|| htmlHidden(html, req, Network)
		|| htmlHidden(html, req, Context)
		|| khtml_attr(
			html, KELEM_INPUT,
			KATTR_TYPE, "date",
			KATTR_NAME, Keys[After].name,
			KATTR_VALUE, date,
			KATTR__MAX
		)
		|| htmlHidden(html, req, Tidy)
		|| khtml_elem(html, KELEM_BUTTON)
		|| khtml_puts(html, "Jump")
		|| khtml_closeelem(html, 2);
}

static enum kcgi_err
displayForm(struct khtmlreq *html, struct kreq *req, bool tidy) {
	if (req->fieldmap[Export]) return KCGI_OK;
	return 0
		|| khtml_attr(
			html, KELEM_FORM,
			KATTR_METHOD, "get",
			KATTR_ACTION, Pages[Events],
			KATTR__MAX
		)
		|| htmlHidden(html, req, Network)
		|| htmlHidden(html, req, Context)
		|| htmlHidden(html, req, After)
		|| htmlHidden(html, req, Before)
		|| khtml_elem(html, KELEM_LABEL)
		|| khtml_attr(
			html, KELEM_INPUT,
			KATTR_TYPE, "checkbox",
			KATTR_NAME, Keys[Tidy].name,
			KATTR_VALUE, "1",
			(tidy ? KATTR_CHECKED : KATTR__MAX), "checked",
			KATTR__MAX
		)
		|| khtml_puts(html, "Hide general events")
		|| khtml_closeelem(html, 1)
		|| khtml_elem(html, KELEM_BUTTON)
		|| khtml_puts(html, "Apply")
		|| khtml_closeelem(html, 2);
}

static char *pageURL(
	const char *network, const char *context,
	enum Key key, time_t time, bool tidy
) {
	char *url = khttp_urlpart(
		NULL, NULL, Pages[Events],
		Keys[Network].name, network,
		Keys[Context].name, context,
		Keys[key].name, timestamp(time),
		(tidy ? Keys[Tidy].name : NULL), "1",
		NULL
	);
	if (!url) err(EX_OSERR, "khttp_urlpart");
	return url;
}

const char *EventsTopicQuery = SQL(
	SELECT topics.topic
	FROM topics
	JOIN contexts USING (context)
	WHERE contexts.network = :network
		AND contexts.name = :context
		AND contexts.query <= NOT :public
		AND topics.time <= strftime('%s', :time)
	ORDER BY topics.time DESC
	LIMIT 1;
);

const char *EventsAfterQuery = SQL(
	SELECT
		events.event,
		events.time,
		events.type,
		names.nick,
		names.user,
		names.host,
		events.target,
		events.message
	FROM events
	JOIN contexts USING (context)
	JOIN names USING (name)
	WHERE contexts.network = :network
		AND contexts.name = :context
		AND contexts.query <= NOT :public
		AND events.type < :tidy
		AND events.time >= strftime('%s', :time)
	ORDER BY events.time, events.event
	LIMIT :limit;
);

const char *EventsBeforeQuery = SQL(
	WITH before AS (
		SELECT
			events.event,
			events.time,
			events.type,
			names.nick,
			names.user,
			names.host,
			events.target,
			events.message
		FROM events
		JOIN contexts USING (context)
		JOIN names USING (name)
		WHERE contexts.network = :network
			AND contexts.name = :context
			AND contexts.query <= NOT :public
			AND events.type < :tidy
			AND events.time < strftime('%s', :time)
		ORDER BY events.time DESC
		LIMIT :limit
	)
	SELECT * FROM before
	ORDER BY time, event;
);

enum kcgi_err eventsPage(struct kreq *req) {
	if (!req->fieldmap[Network] || !req->fieldmap[Context]) {
		return httpFail(req, KHTTP_400);
	}
	const char *network = req->fieldmap[Network]->parsed.s;
	const char *context = req->fieldmap[Context]->parsed.s;

	if (!req->fieldmap[After] && !req->fieldmap[Before]) {
		char *url = khttp_urlpart(
			NULL, NULL, Pages[Events],
			Keys[Network].name, network,
			Keys[Context].name, context,
			Keys[Before].name, timestamp(time(NULL)),
			NULL
		);
		if (!url) err(EX_OSERR, "khttp_urlpart");
		enum kcgi_err error = httpRedirect(req, url);
		free(url);
		return error;
	}

	const char *time = req->fieldmap[Before]
		? req->fieldmap[Before]->parsed.s
		: req->fieldmap[After]->parsed.s;
	bool tidy = (req->fieldmap[Tidy] ? req->fieldmap[Tidy]->parsed.i : false);

	enum kcgi_err error = 0
		|| httpHead(req, KHTTP_200, KMIME_TEXT_HTML)
		|| khttp_body(req);
	if (req->method == KMETHOD_HEAD) return error;

	struct khtmlreq html;
	error = error
		|| khtml_open(&html, req, 0)
		|| htmlHead(&html, req, context)
		|| htmlNav(&html, req)
		|| khtml_elem(&html, KELEM_DIV)
		|| dateForm(&html, req, time)
		|| displayForm(&html, req, tidy)
		|| khtml_closeelem(&html, 1)
		|| khtml_elem(&html, KELEM_TABLE);
	if (error) return error;

	sqlite3_reset(stmt.topic);
	dbBindText(stmt.topic, ":network", network);
	dbBindText(stmt.topic, ":context", context);
	dbBindText(stmt.topic, ":time", time);
	dbBindInt(stmt.topic, ":public", contextsPublic);

	int result = sqlite3_step(stmt.topic);
	if (result == SQLITE_ROW) {
		error = 0
			|| khtml_elem(&html, KELEM_THEAD)
			|| khtml_attr(&html, KELEM_TR, KATTR_CLASS, "topic", KATTR__MAX)
			|| khtml_attr(&html, KELEM_TH, KATTR_COLSPAN, "3", KATTR__MAX)
			|| htmlIRC(&html, sqlite3_column_text(stmt.topic, 0))
			|| khtml_closeelem(&html, 3);
		if (error) return error;
		result = sqlite3_step(stmt.topic);
	}
	if (result != SQLITE_DONE) errx(EX_SOFTWARE, "%s", sqlite3_errmsg(db));

	sqlite3_stmt *events = stmt.eventsAfter;
	if (req->fieldmap[Before]) events = stmt.eventsBefore;

	sqlite3_reset(events);
	dbBindText(events, ":network", network);
	dbBindText(events, ":context", context);
	dbBindText(events, ":time", time);
	dbBindInt(events, ":tidy", (tidy ? Join : TypesLen));
	dbBindInt(events, ":public", contextsPublic);
	dbBindInt(events, ":limit", eventsLimit);

	int rows;
	time_t prevTime = 0;
	for (rows = 0; SQLITE_ROW == (result = sqlite3_step(events)); ++rows) {
		int i = 0;
		struct Event event = {0};
		event.event = sqlite3_column_int64(events, i++);
		event.time = sqlite3_column_int64(events, i++);
		event.type = sqlite3_column_int(events, i++);
		event.nick = sqlite3_column_text(events, i++);
		event.user = sqlite3_column_text(events, i++);
		event.host = sqlite3_column_text(events, i++);
		event.target = sqlite3_column_text(events, i++);
		event.message = sqlite3_column_text(events, i++);

		if (!rows && !req->fieldmap[Export]) {
			char *href = NULL;
			char *page = pageURL(
				network, context, Before, event.time + eventsOverlap, tidy
			);
			asprintf(&href, "%s#%" PRId64, page, event.event);
			if (!href) err(EX_OSERR, "asprintf");
			error = 0
				|| khtml_attr(&html, KELEM_TR, KATTR_CLASS, "page", KATTR__MAX)
				|| khtml_attr(&html, KELEM_TH, KATTR_COLSPAN, "3", KATTR__MAX)
				|| khtml_attr(&html, KELEM_A, KATTR_HREF, href, KATTR__MAX)
				|| khtml_puts(&html, "Earlier messages")
				|| khtml_closeelem(&html, 3);
			free(href);
			free(page);
			if (error) return error;
		}

		if (prevTime && event.time - prevTime >= eventsGap) {
			error = 0
				|| khtml_attr(&html, KELEM_TR, KATTR_CLASS, "gap", KATTR__MAX)
				|| khtml_attr(&html, KELEM_TH, KATTR_COLSPAN, "3", KATTR__MAX)
				|| khtml_puts(&html, "...")
				|| khtml_closeelem(&html, 2);
			if (error) return error;
		}
		prevTime = event.time;

		error = htmlEvent(&html, req, &event);
		if (error) return error;
	}
	if (result != SQLITE_DONE) errx(EX_SOFTWARE, "%s", sqlite3_errmsg(db));

	if ((rows == eventsLimit || req->fieldmap[Before]) && !req->fieldmap[Export]) {
		char *href = pageURL(
			network, context, After, prevTime - eventsOverlap, tidy
		);
		error = 0
			|| khtml_attr(&html, KELEM_TR, KATTR_CLASS, "page", KATTR__MAX)
			|| khtml_attr(&html, KELEM_TH, KATTR_COLSPAN, "3", KATTR__MAX)
			|| khtml_attr(&html, KELEM_A, KATTR_HREF, href, KATTR__MAX)
			|| khtml_puts(&html, "Later messages")
			|| khtml_closeelem(&html, 3);
		free(href);
		if (error) return error;
	}

	if (!rows) {
		error = 0
			|| khtml_elem(&html, KELEM_TR)
			|| khtml_attr(&html, KELEM_TH, KATTR_COLSPAN, "3", KATTR__MAX)
			|| khtml_puts(&html, "No matching messages");
		if (error) return error;
	}

	return htmlFooter(&html, req) || khtml_close(&html);
}
