/* Copyright (C) 2020  C. McEnroe <june@causal.agency>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <err.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>

#include "server.h"

const char *NetworksQuery = SQL(
	WITH recentEvents AS (
		SELECT time, context
		FROM events
		ORDER BY event DESC
		LIMIT :recent
	), activeNetworks AS (
		SELECT network
		FROM contexts
		JOIN recentEvents USING (context)
		WHERE query <= NOT :public
		GROUP BY network
		ORDER BY max(time) DESC
	), allNetworks AS (
		SELECT DISTINCT network
		FROM contexts
		WHERE query <= NOT :public
		ORDER BY network COLLATE NOCASE
	)
	SELECT network, 1 FROM activeNetworks
	UNION ALL
	SELECT network, 0 FROM allNetworks;
);

enum kcgi_err networksPage(struct kreq *req) {
	enum kcgi_err error = 0
		|| httpHead(req, KHTTP_200, KMIME_TEXT_HTML)
		|| khttp_body(req);
	if (req->method == KMETHOD_HEAD) return error;

	struct khtmlreq html;
	error = error
		|| khtml_open(&html, req, 0)
		|| htmlHead(&html, req, "Litterbox")
		|| htmlNav(&html, req);
	if (error) return error;

	sqlite3_reset(stmt.networks);
	dbBindInt(stmt.networks, ":recent", contextsRecent);
	dbBindInt(stmt.networks, ":public", contextsPublic);

	enum State {
		None,
		Active,
		Networks,
	} state = None;
	const char *Headings[] = { NULL, "Active", "Networks" };

	int result;
	while (SQLITE_ROW == (result = sqlite3_step(stmt.networks))) {
		int i = 0;
		const char *network = sqlite3_column_text(stmt.networks, i++);
		bool active = sqlite3_column_int(stmt.networks, i++);

		enum State prev = state;
		state = (active ? Active : Networks);
		if (state != prev) {
			error = 0
				|| khtml_closeelem(&html, 1)
				|| khtml_elem(&html, KELEM_H2)
				|| khtml_puts(&html, Headings[state])
				|| khtml_closeelem(&html, 1)
				|| khtml_elem(&html, KELEM_UL);
			if (error) return error;
		}

		char *href = khttp_urlpart(
			NULL, NULL, Pages[Contexts],
			Keys[Network].name, network,
			NULL
		);
		if (!href) err(EX_OSERR, "khttp_urlpart");
		error = 0
			|| khtml_elem(&html, KELEM_LI)
			|| khtml_attr(&html, KELEM_A, KATTR_HREF, href, KATTR__MAX)
			|| khtml_puts(&html, network)
			|| khtml_closeelem(&html, 2);
		free(href);
		if (error) return error;
	}
	if (result != SQLITE_DONE) errx(EX_SOFTWARE, "%s", sqlite3_errmsg(db));
	return htmlFooter(&html, req) || khtml_close(&html);
}
